#!/usr/bin/env python3
#
# Copyright (c) 2019 Roberto Riggio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Exposes a RESTful interface ."""

import uuid

import empower_core.apimanager.apimanager as apimanager


# pylint: disable=W0223
class WorkersHandler(apimanager.APIHandler):
    """Workers handler."""

    URLS = [r"/api/v1/workers/?",
            r"/api/v1/workers/([a-zA-Z0-9-]*)/?"]

    @apimanager.validate(min_args=0, max_args=1)
    def get(self, worker_id=None):
        """List the workers.

        Args:

            [0]: the worker id (optional)

        Example URLs:

            GET /api/v1/workers

            [
                {
                    "name":
                        "empower.workers.wifichannelstats.wifichannelstats",
                    "params": {
                        "every": 2000,
                        "project_id": "4cd2bca2-8c28-4e66-9c8a-7cbd1ba4e6f9",
                        "service_id": "0f91e8ad-1c2a-4b06-97f9-e34097c4c1d0"
                    }
                }
            ]

            GET /api/v1/workers/0f91e8ad-1c2a-4b06-97f9-e34097c4c1d0

            {
                "name": "empower.workers.wifichannelstats.wifichannelstats",
                "params": {
                    "every": 2000,
                    "project_id": "4cd2bca2-8c28-4e66-9c8a-7cbd1ba4e6f9",
                    "service_id": "0f91e8ad-1c2a-4b06-97f9-e34097c4c1d0"
                }
            }
        """

        return self.service.env.services[uuid.UUID(worker_id)] \
            if worker_id else self.service.env.services

    @apimanager.validate(returncode=201, min_args=0, max_args=1)
    def post(self, *args, **kwargs):
        """Start a new worker.

        Args:

            [0], the service id (optional)

        Request:

            version: protocol version (1.0)
            name: the name of the worker (mandatory)
            params: the list of parmeters to be set (optional)

        Example URLs:

            POST /api/v1/workers
            {
                "version": "1.0",
                "name": "empower.workers.wifichannelstats.wifichannelstats",
                "params": {
                    "every": 5000
                }
            }
        """

        service_id = uuid.UUID(args[0]) if args else None

        params = kwargs['params'] if 'params' in kwargs else {}

        service = \
            self.service.env.register_service(name=kwargs['name'],
                                              params=params,
                                              service_id=service_id)

        self.set_header("Location", "/api/v1/workers/%s" % service.service_id)

    @apimanager.validate(returncode=204, min_args=1, max_args=1)
    def put(self, *args, **kwargs):
        """Update the configuration of a worker.

        Args:

            [0], the worker id (mandatory)

        Request:

            version: protocol version (1.0)
            params: the list of parmeters to be set (optional)

        Example URLs:

            PUT /api/v1/workers/08e14f40-6ebf-47a0-8baa-11d7f44cc228
            {
                "version": "1.0",
                "params":
                {
                    "every": 5000
                }
            }
        """

        service_id = uuid.UUID(args[0])
        params = kwargs['params'] if 'params' in kwargs else {}

        self.service.env.reconfigure_service(service_id=service_id,
                                             params=params)

    @apimanager.validate(returncode=204, min_args=1, max_args=1)
    def delete(self, worker_id):
        """Stop a worker.

        Args:

            [0], the worker id

        Example URLs:

            DELETE /api/v1/workers/08e14f40-6ebf-47a0-8baa-11d7f44cc228
        """

        self.service.env.unregister_service(service_id=uuid.UUID(worker_id))



#!/usr/bin/env python3
#
# Copyright (c) 2019 Roberto Riggio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Ethernet address."""

from pymodm.errors import ValidationError
from pymodm.base.fields import MongoBaseField
from empower_core.serialize import serializable_string


@serializable_string
class EtherAddress:
    """Ethernet address."""

    def __init__(self, addr=None):

        if not addr:
            addr = "00:00:00:00:00:00"

        # Always stores as a 6 character string
        if isinstance(addr, bytes) and len(addr) == 6:
            # raw
            self._value = addr
        elif isinstance(addr, str):
            if len(addr) == 17 or addr.count(':') == 5:
                # hex
                if len(addr) == 17:
                    if addr[2::3] != ':::::' and addr[2::3] != '-----':
                        raise ValueError("Bad format for ethernet address")
                    # Address of form xx:xx:xx:xx:xx:xx
                    # Pick out the hex digits only
                    addr = ''.join(
                        (addr[x * 3:x * 3 + 2] for x in range(0, 6)))
                else:
                    # Assume it's hex digits but they may not all be in
                    # two-digit groupings (e.g., xx:x:x:xx:x:x). This actually
                    # comes up.
                    addr = ''.join(["%02x" % (int(x, 16),)
                                    for x in addr.split(":")])

                # We should now have 12 hex digits (xxxxxxxxxxxx).
                # Convert to 6 raw bytes.
                addr = b''.join(bytes((int(addr[x * 2:x * 2 + 2], 16),))
                                for x in range(0, 6))
            else:
                raise ValueError("Expected 6 raw bytes or some hex")
            self._value = addr
        elif isinstance(addr, EtherAddress):
            self._value = addr.to_raw()
        elif addr is None:
            self._value = b'\x00' * 6
        else:
            raise ValueError("EtherAddress must be a string of 6 raw bytes")

    def is_global(self):
        """Returns True if this is a globally unique (OUI enforced) address."""
        return not self.is_local()

    def is_local(self):
        """Returns True if this is a locally-administered address."""
        return bool(self._value[0] & 2)

    def is_multicast(self):
        """Returns True if this is a multicast address."""

        return bool(self._value[0] & 1)

    def to_raw(self):
        """Returns the address as a 6-long bytes object."""

        return self._value

    def to_str(self, separator=':'):
        """Return an ASCII representation of the object."""

        return separator.join(('%02x' % (x,) for x in self._value)).upper()

    def match(self, other):
        """ Bitwise match. """

        if isinstance(other, EtherAddress):
            other = other.to_raw()
        elif isinstance(other, bytes):
            pass
        else:
            try:
                other = EtherAddress(other).to_raw()
            except ValueError:
                return False
        for cnt in range(0, 6):
            if (self._value[cnt] & other[cnt]) != self._value[cnt]:
                return False
        return True

    def __str__(self):
        return self.to_str()

    def __eq__(self, other):

        if isinstance(other, EtherAddress):
            other = other.to_raw()
        elif isinstance(other, bytes):
            pass
        else:
            try:
                other = EtherAddress(other).to_raw()
            except ValueError:
                return False
        if self._value == other:
            return True
        return False

    def __hash__(self):
        return self._value.__hash__()

    def __repr__(self):
        return self.__class__.__name__ + "('" + self.to_str() + "')"

    def __setattr__(self, a, v):
        if hasattr(self, '_value'):
            raise TypeError("This object is immutable")
        object.__setattr__(self, a, v)

    @classmethod
    def bcast(cls):
        """ Return a broadcast address. """

        return EtherAddress('ff:ff:ff:ff:ff:ff')


class EtherAddressField(MongoBaseField):
    """A field that stores EtherAddresses."""

    def __init__(self, verbose_name=None, mongo_name=None, **kwargs):

        super(EtherAddressField, self).__init__(verbose_name=verbose_name,
                                                mongo_name=mongo_name,
                                                **kwargs)

        def validate_ethernet_address(value):

            try:
                EtherAddress(value)
            except ValueError:
                msg = '%r is not a valid Ethernet address.' % value
                raise ValidationError(msg)

        self.validators.append(validate_ethernet_address)

    @classmethod
    def to_mongo(cls, value):
        """Convert value for storage."""

        try:
            return str(value)
        except ValueError:
            msg = '%r is not a valid Ethernet address.' % value
            raise ValidationError(msg)

    @classmethod
    def to_python(cls, value):
        """Convert value back to Python."""

        try:
            return EtherAddress(value)
        except ValueError:
            msg = '%r is not a valid Ethernet address.' % value
            raise ValidationError(msg)



#!/usr/bin/env python3
#
# Copyright (c) 2019 Roberto Riggio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Command Line Interface."""

import sys
import base64
import getpass
import json
import pkgutil
import importlib

from argparse import ArgumentParser

import requests

from requests.exceptions import ConnectionError as RequestConnectionError

import empower_core.cli as help_commands

from empower_core.serialize import serialize

USAGE = "%(prog)s {0}"
URL = "%s://%s%s:%s"

DESCS = {}
CMDS = {}


def load_commands(commands):
    """Load commands from specified package."""

    for _, name, is_pkg in pkgutil.walk_packages(commands.__path__):

        if is_pkg and name.endswith("_commands"):

            package = importlib.import_module(commands.__name__ + "." + name)
            pkgs = pkgutil.walk_packages(package.__path__)

            for _, module_name, _ in pkgs:

                module = importlib.import_module(package.__name__ + "." +
                                                 module_name)

                cmd_name = module_name.replace("_", "-")
                pa_cmd = module.pa_cmd if hasattr(module, 'pa_cmd') else None
                do_cmd = module.do_cmd

                CMDS[cmd_name] = [pa_cmd, do_cmd]
                DESCS[cmd_name] = module.__doc__


def connect(gargs, cmd, expected=200, request=None,
            headers=None, exit_on_err=True):
    """ Run command. """

    if not headers:
        headers = get_headers(gargs)

    url = "%s://%s:%s" % (gargs.transport, gargs.host, gargs.port)
    method = getattr(requests, cmd[0].lower())

    if request:
        request["version"] = "1.0"

    response = method(url + cmd[1], headers=headers, json=serialize(request))

    try:
        data = json.loads(response.text)
    except json.decoder.JSONDecodeError:
        data = {}

    if response.status_code != expected:

        msg = "%u: %s (%s)" % \
            (data['status_code'], data['title'], data['detail'])

        if exit_on_err:
            print(msg)
            sys.exit()
        else:
            raise Exception(msg)

    return response, data


def get_headers(gargs):
    """Get the required headers. """

    headers = {
        'Content-type': 'application/json'
    }

    if gargs.no_passwd:
        return headers
    if gargs.passwdfile is not None:
        passwd = open(gargs.passwdfile, "r").read().strip()
        auth_str = "%s:%s" % (gargs.user, passwd)
    elif gargs.auth is not None:
        auth_str = gargs.auth
    else:
        passwd = getpass.getpass("Password: ")
        auth_str = "%s:%s" % (gargs.user, passwd)

    auth = base64.b64encode(auth_str.encode('utf-8'))

    headers['Authorization'] = 'Basic %s' % auth.decode('utf-8')

    return headers


def get_params(leftovers):
    """Turn arguments leftovers into service params."""

    params = {}

    for param in leftovers:
        tokens = param.split("=")
        if len(tokens) != 2:
            continue
        key = tokens[0].replace("--", "")
        value = tokens[1]
        params[key] = value

    return params


def pa_none(args, cmd):
    """ Null parser method. """

    parser = ArgumentParser(usage=USAGE.format(cmd), description=DESCS[cmd])
    (args, leftovers) = parser.parse_known_args(args)
    return args, leftovers


def parse_global_args(arglist):
    """ Parse global arguments list. """

    usage = "%s [options] command [command options]" % sys.argv[0]

    args = []

    while arglist and arglist[0] not in CMDS:
        args.append(arglist[0])
        arglist.pop(0)

    parser = ArgumentParser(usage=usage)

    parser.add_argument("-r", "--host", dest="host", default="127.0.0.1",
                        help="REST server address; default='127.0.0.1'")
    parser.add_argument("-p", "--port", dest="port", default="8888",
                        help="REST server port; default=8888")
    parser.add_argument("-u", "--user", dest="user", default="root",
                        help="EmPOWER admin user; default='root'")
    parser.add_argument("-a", "--auth", dest="auth", default=None,
                        help="EmPOWER admin user:passw")
    parser.add_argument("-n", "--no-passwd", action="store_true",
                        dest="no_passwd", default=False,
                        help="Run without password; default false")
    parser.add_argument("-f", "--passwd-file", dest="passwdfile",
                        default=None, help="Password file; default=none")
    parser.add_argument("-t", "--transport", dest="transport", default="http",
                        help="The transport (http/https); default='http'")

    (args, _) = parser.parse_known_args(args)

    return args, arglist, parser


def print_available_cmds():
    """ Print list of available commands. """

    cmds = list(CMDS.keys())
    if 'help' in cmds:
        cmds.remove('help')
    cmds.sort()
    print("\nAvailable commands are: ")
    for cmd in cmds:
        print("   {0:25}     {1:10}".format(cmd, DESCS[cmd]))
    print("\nSee '%s help <command>' for more info." % sys.argv[0])


def main(commands=None):
    """ Parse argument list and execute command. """

    load_commands(help_commands)

    if commands:
        load_commands(commands)

    (gargs, rargs, parser) = parse_global_args(sys.argv[1:])

    if len(sys.argv) == 1:
        print(parser.format_help().strip())
        print_available_cmds()
        sys.exit()

    if not rargs:
        print("Unknown command")
        print_available_cmds()
        sys.exit()

    (parse_args, do_func) = CMDS[rargs[0]]

    if parse_args:
        (args, leftovers) = parse_args(rargs[1:], rargs[0])
    else:
        (args, leftovers) = pa_none(rargs[1:], rargs[0])

    if leftovers and rargs[0] != "help":
        print("Warning - unknown parameters: ", ', '.join(leftovers))

    try:
        do_func(gargs, args, leftovers)
    except RequestConnectionError:
        print("Failed to establish a connection with the Controller")


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
#
# Copyright (c) 2020 Roberto Riggio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Exposes a RESTful interface ."""

import empower_core.apimanager.apimanager as apimanager


# pylint: disable=W0223
class CatalogHandler(apimanager.APIHandler):
    """Access the workers catalog."""

    URLS = [r"/api/v1/catalog/?",
            r"/api/v1/catalog/(.*)/?"]

    @apimanager.validate(min_args=0, max_args=1)
    def get(self, name=None):
        """List of available workers.

        Example URLs:

             GET /api/v1/catalog

            [
                {
                    "name":
                        "empower.workers.wifichannelstats.wifichannelstats",
                    "params": {
                        "every": 2000,
                        "project_id": "4cd2bca2-8c28-4e66-9c8a-7cbd1ba4e6f9",
                        "service_id": "0f91e8ad-1c2a-4b06-97f9-e34097c4c1d0"
                    }
                }
            ]
        """

        return self.service.catalog[name] if name else self.service.catalog



#!/usr/bin/env python3
#
# Copyright (c) 2019 Roberto Riggio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""EUI64 data format."""

import re

from pymodm.errors import ValidationError
from pymodm.base.fields import MongoBaseField

from empower_core.serialize import serializable_string

EUI64_PATTERN = re.compile("^([0-9a-fA-F]){2}(:([0-9a-fA-F]){2}){7}$")
"""Regular expression matching a well formed EUI using colon """

PLAIN_PATTERN = re.compile("^([0-9a-fA-F]){16}$")
"""Regular expression matching a well formed EUI no colon """

ID6_PATTERN = re.compile("^([0-9a-fA-F]){1,4}(:([0-9a-fA-F]){1,4}){3}$")
"""Id6 regular expression matching representation after :: expansion"""


@serializable_string
class EUI64():
    """EUI64."""

    DEFAULT_VALUE = None
    BYTES = 8

    def __init__(self, data=None):
        """Initialize internal value of EUI64."""
        if data is None:
            # set to default
            self._value = self.DEFAULT_VALUE
        else:
            self._value = self._to_raw_(data)

    @property
    def id6(self):
        """Return data in ID6 format."""
        if self._value == 0:
            return "::0"
        out = "{0:#0{1}x}".format(self._value, self.BYTES*2+2)[2:]
        out = [out[4*i:4*i+4]+':' for i in range(int(len(out)/4))]
        out = ''.join(out)[:-1]
        return out

    @property
    def eui64(self):
        """Return data in EUI format."""
        if self._value == 0:
            return "00:00:00:00:00:00:00:00"
        out = "{0:#0{1}x}".format(self._value, self.BYTES*2 + 2)[2:]
        out = [out[2*i:2*i+2]+':' for i in range(int(len(out)/2))]
        out = ''.join(out)[:-1]
        return out

    @property
    def hex(self):
        """Return EUID in hex str format."""
        return hex(self._value)[2:]

    @classmethod
    def _to_raw_(cls, data):
        """Return the data as internal representation."""
        if isinstance(data, int):
            return data
        if isinstance(data, EUI64):
            return int(data)
        if isinstance(data, bytes) and len(data) == cls.BYTES*2:
            # raw data
            return int.from_bytes(data, byteorder="little")
        if not isinstance(data, str):
            raise ValueError('Invalid EUID: %r' % data)
        data = re.sub(r' ', '', data)
        # eliminate spaces
        data = re.sub(r'[.-]', ':', data)
        # substitute "." and "-" with ":"
        if EUI64_PATTERN.match(data) or PLAIN_PATTERN.match(data):
            data = re.sub(r':', '', data)
            data = [data[i:i+2] for i in range(0, len(data), 2)]
            return ((int(data[0], 16) << 56) |
                    (int(data[1], 16) << 48) |
                    (int(data[2], 16) << 40) |
                    (int(data[3], 16) << 32) |
                    (int(data[4], 16) << 24) |
                    (int(data[5], 16) << 16) |
                    (int(data[6], 16) << 8) |
                    (int(data[7], 16)))
        # first check if an :: expansion is needed in ID6
        if data in ["0", "::"]:
            data = "0:0:0:0"
        if data.count(':') == 2:
            data = re.sub('^::', '0:0:0:', data)
            data = re.sub('::$', ':0:0:0', data)
            data = re.sub('::', ':0:0:', data)
        elif data.count(':') == 3:
            data = re.sub('^::', '0:0:', data)
            data = re.sub('::$', ':0:0', data)
            data = re.sub('::', ':0:', data)
        if ID6_PATTERN.match(data):
            data = data.split(':')
            return ((int(data[0], 16) << 48) |
                    (int(data[1], 16) << 32) |
                    (int(data[2], 16) << 16) |
                    (int(data[3], 16)))
        raise ValueError('Invalid EUID: %r' % data)

    def __eq__(self, other):
        """Return if two istances are equal."""
        if isinstance(other, EUI64):
            return self._value == int(other)
        return False

    def __ne__(self, other):
        """Return if two istances are not equal."""
        return not self.__eq__(other)

    def __repr__(self):
        """Return a representation of the instance."""
        return self.__class__.__name__ + "('" + self.id6 + "')"

    def __bool__(self):
        """Return the bool value."""
        return bool(self._value)

    def __hash__(self):
        """Return a hash value."""
        return hash(self._value)

    def __bytes__(self, byteorder="little"):
        """Return EUID in bytes format."""
        byteorder = "little"
        return self._value.to_bytes(8, byteorder)

    def __str__(self):
        """Return EUID in string format."""
        return self.id6

    def __int__(self):
        """Return EUID in int format."""
        return self._value


class EUI64Field(MongoBaseField):
    """A field that stores WS URIs.

    This field only accepts EUID64 or ID6 str.
    """

    def __init__(self, verbose_name=None, mongo_name=None, **kwargs):
        """Init EUID data.

        :parameters:
          - "verbose_name": A human-readable name for the Field.
          - "mongo_name": The name of this field when stored in MongoDB.
        """
        super().__init__(verbose_name=verbose_name,
                         mongo_name=mongo_name,
                         **kwargs)

    @classmethod
    def validate(cls, value):
        """Convert value in id6 format."""
        if isinstance(value, EUI64):
            return value.id6
        return EUI64(value).id6

    @classmethod
    def to_mongo(cls, value):
        """Convert EUI64 to a value to be stored in mongodb."""
        if isinstance(value, EUI64):
            return value.id6
        try:
            return EUI64(value).id6
        except ValueError:
            msg = "%r is not a valid EUID." % value
            raise ValidationError(msg)

    @classmethod
    def to_python(cls, value):
        """Convert a value stored in mongodb to EUI64."""
        try:
            return EUI64(value)
        except ValidationError:
            return value



#!/usr/bin/env python3
#
# Copyright (c) 2019 Roberto Riggio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Inspect the specified module for services."""

import pkgutil
import importlib


def walk_module(package):
    """Inspect the specified module for services."""

    package = importlib.import_module(package)

    results = {}

    pkgs = pkgutil.walk_packages(package.__path__)

    for importer, module_name, is_pkg in pkgs:

        if not is_pkg:
            continue

        # only import from subfolders of the package
        if not any(importer.path.startswith(p) for p in package.__path__):
            continue

        __import__(package.__name__ + "." + module_name)

        if not hasattr(package, module_name):
            continue

        module = getattr(package, module_name)

        if not hasattr(module, "MANIFEST"):
            continue

        manifest = getattr(module, "MANIFEST")

        if 'name' not in manifest:
            name = package.__name__ + "." + module_name + "." + module_name
            manifest['name'] = name

        if 'modules' not in manifest:
            manifest['modules'] = []

        if 'label' not in manifest:
            manifest['label'] = module_name

        if 'desc' not in manifest:
            manifest['desc'] = "No description available"

        results[manifest['name']] = manifest

    return results



#!/usr/bin/env python3
#
# Copyright (c) 2019 Roberto Riggio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Runtime configuration."""

import uuid
import platform

from importlib import import_module
from pymodm import MongoModel, fields

import empower_core.serialize as serialize

from empower_core.launcher import srv_or_die
from empower_core.serialize import serializable_dict
from empower_core.worker import EWorker


@serializable_dict
class Env(MongoModel):
    """Env class

    Attributes:
        project_id: The project identifier. Notice how there is only one Env
            instance active at any given time and that the project_id of the
            workers loaded by Env is always set to the Env's project_id
        bootstrap: the list of services to be loaded at bootstrap
        services: the services
    """

    project_id = fields.UUIDField(primary_key=True)
    bootstrap = fields.DictField(required=False, blank=True)

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        # List of services in this Env/Project
        self.services = {}

        # Save pointer to EnvManager
        self.manager = srv_or_die("envmanager")

    @classmethod
    def write_points(cls, points):
        """Write points to time-series manager."""

        ts_manager = srv_or_die("tsmanager")
        ts_manager.write_points(points)

    def save_service_state(self, service_id):
        """Save service state."""

        service = self.services[service_id]

        self.bootstrap[str(service.service_id)] = {
            "name": service.name,
            "params": serialize.serialize(service.params),
            "callbacks": serialize.serialize(service.callbacks),
            "configuration": serialize.serialize(service.configuration)
        }

        self.save()

    def remove_service_state(self, service_id):
        """Remove service state."""

        del self.bootstrap[str(service_id)]

        self.save()

    def load_service(self, service_id, name, params):
        """Load a service instance."""

        init_method = getattr(import_module(name), "launch")
        service = init_method(context=self, service_id=service_id, **params)

        if not isinstance(service, EWorker):
            raise ValueError("Service %s not EWorker type" % name)

        return service

    def register_service(self, name, params, service_id=None):
        """Register service."""

        if not service_id:

            # Load a service object using the specified parameters
            requested = self.load_service(uuid.uuid4(), name, params)

            # Check if a service with the same parameters already exists
            services = self.services.values()

            found = next((s for s in services if s == requested), None)

            if found:
                return found

        # Generate a service_id if necessary
        if not service_id:
            service_id = uuid.uuid4()

        # Start the service
        service = self.start_service(service_id, name, params)

        # Save service state
        self.save_service_state(service.service_id)

        return service

    def unregister_service(self, service_id):
        """Unregister service."""

        # If not found abort
        if service_id not in self.services:
            raise KeyError("Service %s not running" % service_id)

        # Stop the service
        self.stop_service(service_id)

        # Remove service state
        self.remove_service_state(service_id)

    def reconfigure_service(self, service_id, params):
        """Reconfigure service."""

        if service_id not in self.services:
            raise KeyError("Service %s not registered" % service_id)

        service = self.services[service_id]

        for param in params:
            if param not in self.bootstrap[str(service_id)]['params']:
                raise KeyError("Param %s undefined" % param)

            setattr(service, param, params[param])

        self.save_service_state(service_id)

        return self.services[service_id]

    def start_services(self):
        """Start registered services."""

        for service_id in list(self.bootstrap):

            try:

                name = self.bootstrap[service_id]['name']
                params = self.bootstrap[service_id]['params']
                callbacks = self.bootstrap[service_id]['callbacks']
                configuration = self.bootstrap[service_id]['configuration']

                service_id = uuid.UUID(service_id)

                self.start_service(service_id, name, params, configuration,
                                   callbacks)

            except ModuleNotFoundError as ex:

                self.manager.log.error("Unable to start service %s: %s",
                                       name, ex)

                self.remove_service_state(service_id)

            except TypeError as ex:

                self.manager.log.error("Unable to start service %s: %s",
                                       name, ex)

                self.remove_service_state(service_id)

            except ValueError as ex:

                self.manager.log.error("Unable to start service %s: %s",
                                       name, ex)

                self.remove_service_state(service_id)

    def stop_services(self):
        """Stop registered services."""

        for service_id in self.bootstrap:
            self.stop_service(uuid.UUID(service_id))

    def start_service(self, service_id, name, params, configuration=None,
                      callbacks=None):
        """Start a service."""

        # wait we are trying to start a service that already exists, abort
        if service_id in self.services:
            raise ValueError("Service %s is already running" % service_id)

        # this will look for the launch method and call it
        self.manager.log.info("Loading service: %s (%s)", name, service_id)
        self.manager.log.info(" - params: %s", params)

        service = self.load_service(service_id, name, params)

        if not service:
            self.manager.log.error("Unable to start service id %s name %s",
                                   service_id, name)
            return

        # add to service list
        self.services[service.service_id] = service

        # set configuration
        if configuration:
            for entry in configuration:
                setattr(service, entry, configuration[entry])

        # set callbacks
        if callbacks:
            service.callbacks = callbacks

        # register handlers
        for handler in service.HANDLERS:
            api_manager = srv_or_die("apimanager")
            api_manager.register_handler(handler)
            handler.service = service

        # start service
        self.manager.log.info("Starting service: %s (%s)", name, service_id)
        service.start()

        return service

    def stop_service(self, service_id):
        """Stop a service."""

        if service_id not in self.services:
            raise KeyError("Service %s not running" % service_id)

        self.services[service_id].stop()
        del self.services[service_id]

    def to_dict(self):
        """Return JSON-serializable representation of the object."""

        output = {}

        output['project_id'] = self.project_id
        output['bootstrap'] = self.bootstrap

        output['platform'] = {
            "machine": platform.machine(),
            "node": platform.node(),
            "platform": platform.platform(),
            "processor": platform.processor(),
            "python_version": platform.python_version(),
            "release": platform.release(),
            "system": platform.system(),
            "version": platform.version(),
        }

        return output

    def to_str(self):
        """Return an ASCII representation of the object."""

        return str(self.project_id)

    def __str__(self):
        return self.to_str()

    def __hash__(self):
        return hash(self.project_id)

    def __eq__(self, other):
        if isinstance(other, Env):
            return self.project_id == other.project_id
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return self.__class__.__name__ + "('" + self.to_str() + "')"



#!/usr/bin/env python3
#
# Copyright (c) 2020 Roberto Riggio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Accounts manager."""

from empower_core.service import EService
from empower_core.accountsmanager.account import Account
from empower_core.accountsmanager.accountshandler import AccountsHandler


class AccountsManager(EService):
    """Accounts manager."""

    HANDLERS = [AccountsHandler]

    ACCOUNT_IMPL = Account

    accounts = {}

    def start(self):
        """Start accounts manager."""

        super().start()

        for account in self.ACCOUNT_IMPL.objects.all():
            self.accounts[account.username] = account

        if "root" not in self.accounts:

            self.log.info("No root user found, creating defaults!")

            self.create(username="root",
                        password="root",
                        name="admin",
                        email="admin@5g-empower.io")

            self.create(username="foo",
                        password="foo",
                        name="Foo",
                        email="foo@5g-empower.io")

            self.create(username="bar",
                        password="bar",
                        name="Bar",
                        email="bar@5g-empower.io")

    def check_permission(self, username, password):
        """Check if username/password match."""

        if username not in self.accounts:
            return False

        if self.accounts[username].password != password:
            return False

        return True

    def create(self, username, name, email, password):
        """Create new account."""

        if username in self.accounts:
            raise ValueError("Duplicate username %s found" % username)

        user = self.ACCOUNT_IMPL(username=username,
                                 password=password,
                                 name=name,
                                 email=email)

        user.save()

        self.accounts[username] = user

        return self.accounts[username]

    def update(self, username, name, email, password=None):
        """Update account."""

        if username not in self.accounts:
            raise KeyError("Username %s not found" % username)

        user = self.accounts[username]

        try:
            user.name = name
            user.email = email
            if password:
                user.password = password
            user.save()
        finally:
            user.refresh_from_db()

        return self.accounts[username]

    def remove(self, username):
        """Check if username/password match."""

        if username == "root":
            raise ValueError("The 'root' account cannot be removed")

        if username not in self.accounts:
            raise KeyError("Username %s not found" % username)

        user = self.accounts[username]

        user.delete()

        del self.accounts[username]


def launch(context, service_id):
    """ Initialize the module. """

    return AccountsManager(context=context, service_id=service_id)



#!/usr/bin/env python3
#
# Copyright (c) 2019 Roberto Riggio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Projects manager."""

from empower_core.walkmodule import walk_module
from empower_core.launcher import srv_or_die
from empower_core.service import EService
from empower_core.envmanager.envmanager import EnvManager
from empower_core.projectsmanager.project import Project
from empower_core.projectsmanager.appcallbackhandler import \
    AppCallbacksHandler
from empower_core.projectsmanager.cataloghandler import CatalogHandler
from empower_core.projectsmanager.appshandler import AppsHandler
from empower_core.projectsmanager.projectshandler import ProjectsHandler


class ProjectsManager(EnvManager):
    """Projects manager."""

    HANDLERS = [CatalogHandler, AppsHandler, ProjectsHandler,
                AppCallbacksHandler]

    PROJECT_IMPL = Project

    projects = {}

    def start(self):
        """Start projects manager."""

        super().start()

        for project in self.PROJECT_IMPL.objects.all():
            self.projects[project.project_id] = project
            self.projects[project.project_id].start_services()

    def create(self, desc, project_id, owner):
        """Create new project."""

        if project_id in self.projects:
            raise ValueError("Project %s already defined" % project_id)

        accounts_manager = srv_or_die("accountsmanager")

        if owner not in accounts_manager.accounts:
            raise KeyError("Username %s not found" % owner)

        project = self.PROJECT_IMPL(project_id=project_id,
                                    desc=desc,
                                    owner=owner)

        project.save()

        self.projects[project_id] = project

        self.projects[project_id].start_services()

        return self.projects[project_id]

    def update(self, project_id, desc):
        """Update project."""

        if project_id not in self.projects:
            raise KeyError("Project %s not found" % project_id)

        project = self.projects[project_id]

        try:
            project.desc = desc
            project.save()
        finally:
            project.refresh_from_db()

        return self.projects[project_id]

    def remove_all(self):
        """Remove all projects."""

        for project_id in list(self.projects):
            self.remove(project_id)

    def remove(self, project_id):
        """Remove project."""

        # Check if project exists
        if project_id not in self.projects:
            raise KeyError("Project %s not registered" % project_id)

        # Fetch project
        project = self.projects[project_id]

        # Stop running services
        self.projects[project_id].stop_services()

        # Delete project from datase and manager
        project.delete()
        del self.projects[project_id]


def launch(context, service_id, catalog_packages=""):
    """ Initialize the module. """

    return ProjectsManager(context=context, service_id=service_id,
                           catalog_packages=catalog_packages)



#!/usr/bin/env python3
#
# Copyright (c) 2019 Roberto Riggio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Conf manager."""

import uuid

from empower_core.walkmodule import walk_module
from empower_core.service import EService
from empower_core.envmanager.env import Env
from empower_core.envmanager.workercallbackshandler import \
    WorkerCallbacksHandler
from empower_core.envmanager.workershandler import WorkersHandler
from empower_core.envmanager.cataloghandler import CatalogHandler
from empower_core.envmanager.envhandler import EnvHandler


class EnvManager(EService):
    """Environment manager."""

    HANDLERS = [WorkersHandler, WorkerCallbacksHandler, CatalogHandler,
                EnvHandler]

    ENV_IMPL = Env

    env = None

    @property
    def catalog(self):
        """Return workers_package."""

        return walk_module(self.catalog_packages)

    @property
    def catalog_packages(self):
        """Return catalog_packages."""

        return self.params["catalog_packages"]

    @catalog_packages.setter
    def catalog_packages(self, value):
        """Set catalog_packages."""

        self.params["catalog_packages"] = value

    def start(self):
        """Start configuration manager."""

        super().start()

        if not self.ENV_IMPL.objects.all().count():
            self.ENV_IMPL(project_id=uuid.uuid4()).save()

        self.env = self.ENV_IMPL.objects.first()
        self.env.start_services()


def launch(context, service_id, catalog_packages=""):
    """ Initialize the module. """

    return EnvManager(context=context, service_id=service_id,
                      catalog_packages=catalog_packages)



#!/usr/bin/env python3
#
# Copyright (c) 2019 Roberto Riggio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Public land mobile network identifier (PLMNID)."""

from stdnum import numdb
from stdnum.util import clean, isdigits

from pymodm.errors import ValidationError
from pymodm.base.fields import MongoBaseField

from empower_core.serialize import serializable_string


@serializable_string
class PLMNID:
    """Public land mobile network  identifier (PLMNID)."""

    def __init__(self, plmnid):

        plmnid = clean(plmnid, ' -').strip().upper()

        if not isdigits(plmnid):
            raise ValueError("Invalid PLMNID %s" % plmnid)

        if len(plmnid) not in (4, 5):
            raise ValueError("Invalid PLMNID length %s" % plmnid)

        if len(tuple(numdb.get('imsi').split(plmnid))) < 2:
            raise ValueError("Invalid PLMNID format %s" % plmnid)

        self.info = dict(plmnid=plmnid)

        mcc_info, mnc_info = numdb.get('imsi').info(plmnid)

        self.info['mcc'] = mcc_info[0]
        self.info.update(mcc_info[1])
        self.info['mnc'] = mnc_info[0]
        self.info.update(mnc_info[1])

    @property
    def mcc(self):
        """Get mcc."""

        return self.info['mcc']

    @property
    def mnc(self):
        """Get mnc."""

        return self.info['mnc']

    def to_str(self):
        """Return an ASCII representation of the object."""

        return "%s%s" % (self.mcc, self.mnc)

    def to_tuple(self):
        """Return a tuple representation of the object."""

        return tuple(self.info.values())

    def to_dict(self):
        """Return a dict representation of the object."""

        return self.info

    def __str__(self):
        return self.to_str()

    def __len__(self):
        return len(self.to_str())

    def __hash__(self):
        return hash(self.to_str())

    def __eq__(self, other):
        if isinstance(other, PLMNID):
            return self.to_str() == other.to_str()
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return self.__class__.__name__ + "('" + self.to_str() + "')"


class PLMNIDField(MongoBaseField):
    """A field that stores PLMNIDs."""

    def __init__(self, verbose_name=None, mongo_name=None, **kwargs):

        super(PLMNIDField, self).__init__(verbose_name=verbose_name,
                                          mongo_name=mongo_name,
                                          **kwargs)

        def validate_plmnid(value):

            try:

                if isinstance(value, PLMNID):
                    value = value.to_str()

                return PLMNID(value)

            except ValueError:
                msg = '%r is not a valid PLMNID.' % value
                raise ValidationError(msg)

        self.validators.append(validate_plmnid)

    @classmethod
    def to_mongo(cls, value):
        """Convert value for storage."""

        try:
            return value.to_str()
        except ValueError:
            msg = '%r is not a valid PLMNID.' % value
            raise ValidationError(msg)

    @classmethod
    def to_python(cls, value):
        """Convert value back to Python."""

        try:

            if isinstance(value, PLMNID):
                value = value.to_str()

            return PLMNID(value)

        except ValueError:
            msg = '%r is not a valid PLMNID.' % value
            raise ValidationError(msg)



#!/usr/bin/env python3
#
# Copyright (c) 2019 Roberto Riggio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Service."""

import uuid
import logging
import types

from urllib.parse import urlparse

import requests
import tornado.ioloop

from empower_core.imsi import IMSI
from empower_core.plmnid import PLMNID
from empower_core.etheraddress import EtherAddress
from empower_core.ssid import SSID
from empower_core.serialize import serializable_dict
from empower_core.serialize import serialize

CALLBACK_NATIVE = "native"
CALLBACK_REST = "rest"

CALLBACK_TYPES = [CALLBACK_NATIVE, CALLBACK_REST]


@serializable_dict
class EService:
    """Base service class."""

    HANDLERS = []

    def __init__(self, context, service_id, **kwargs):

        # Declaration
        self._service_id = None

        # Pointer to an Env or a Project object (can be None)
        self.context = context

        # Set the service id
        self.service_id = service_id

        # If every is not specified then disable periodic loop
        if 'every' not in kwargs:
            kwargs['every'] = -1

        # Service's callbacks
        self.callbacks = dict()

        # Set logger
        self.log = logging.getLogger(self.name)

        # Worker process, set only if every > 0
        self.worker = None

        # Persistent dict
        self.configuration = {}

        # Service parameters
        self.params = {}

        # Set service parameters
        for param in kwargs:
            setattr(self, param, kwargs[param])

    def validate_param(self, param, value):
        """Validate the parameter using the manifest."""

        if param not in self.manifest['params']:
            raise ValueError("Invalid paramer: %s" % param)

        manifest = self.manifest['params'][param]
        static = manifest.get('static', False)

        if param in self.params and static:
            raise ValueError("Parameter cannot be modified: %s" % param)

        if not value and not manifest['mandatory']:
            return manifest['default']

        if not value and manifest['mandatory']:
            raise ValueError("Parameter %s cannot be null" % param)

        param_type = manifest['type']

        cast_dict = {
            "str": str,
            "int": int,
            "IMSI": IMSI,
            "PLMNID": PLMNID,
            "SSID": SSID,
            "EtherAdress": EtherAddress
        }

        # try to cast value to specified type
        if isinstance(param_type, str):

            if param_type not in cast_dict:
                raise ValueError("Invalid value for '%s' expected '%s'" %
                                 (param, param_type))

            param_type_class = cast_dict[param_type]
            return param_type_class(value)

        # check if in the list
        if isinstance(param_type, list):

            if value in param_type:
                return value

            raise ValueError("Invalid value for '%s' valid '%s' got '%s'" %
                             (param, ','.join(str(x) for x in param_type),
                              value))

        # value is invalid
        raise ValueError("Invalid value for '%s': %s" % (param, value))

    def save_service_state(self):
        """Save service state."""

        if not self.context:
            return

        self.context.save_service_state(self.service_id)

    def register_service(self, name, **kwargs):
        """Register a service.

        If a service with the same name and parameters is already running, then
        that service instance will be returned. Otherwise a new service will be
        spawned."""

        if not self.context:
            return None

        return self.context.register_service(name, params=kwargs)

    def unregister_service(self, service_id):
        """Unregister a service."""

        if not self.context:
            return None

        return self.context.unregister_service(service_id)

    def handle_callbacks(self, params=None, name="default"):
        """Invoke registered callbacks."""

        if name not in self.callbacks:
            return

        callback_type = self.callbacks[name]['callback_type']
        argument = params if params else self
        callback = self.callbacks[name]['callback']

        self.log.info("Handling callback %s (%s)", name, callback_type)

        try:

            # if type is native just invoke it with the specified params
            if self.callbacks[name]['callback_type'] == CALLBACK_NATIVE:
                callback(argument)
                return

            # if type is REST make a post with the serialized params
            if self.callbacks[name]['callback_type'] == CALLBACK_REST:
                response = requests.post(url=callback,
                                         json=serialize(argument))
                self.log.info("POST %s - %u", callback, response.status_code)

        except Exception as ex:
            self.log.exception(ex)

    def add_callback(self, callback, name="default",
                     callback_type=CALLBACK_NATIVE):
        """Add a new callback."""

        if 'callbacks' not in self.manifest:
            raise KeyError("Callback %s not defined" % name)

        if name not in self.manifest['callbacks']:
            raise KeyError("Callback %s not defined" % name)

        if callback_type not in CALLBACK_TYPES:
            raise ValueError("Invalid callback type: %s" % callback_type)

        # native callbacks are not save to the service state
        if callback_type == CALLBACK_NATIVE:

            if not isinstance(callback,
                              (types.FunctionType, types.MethodType)):

                raise ValueError("Callback not callable")

            self.callbacks[name] = {
                "name": name,
                "callback": callback,
                "callback_type": CALLBACK_NATIVE
            }

            return

        # rest callbacks can be saved to service state
        if callback_type == CALLBACK_REST:

            self.callbacks[name] = {
                "name": name,
                "callback": urlparse(callback).geturl(),
                "callback_type": CALLBACK_REST
            }

            self.save_service_state()

            return

        raise ValueError("Invalid input to add callback")

    def rem_callback(self, name="default"):
        """Remove a callback."""

        del self.callbacks[name]
        self.save_service_state()

    def to_dict(self):
        """Return JSON-serializable representation of the object."""

        output = {}

        output['service_id'] = self.service_id
        output['manifest'] = self.manifest
        output['name'] = self.name
        output['callbacks'] = self.callbacks
        output['params'] = self.params

        if self.context:
            output['project_id'] = self.context.project_id

        return output

    @property
    def name(self):
        """Get name."""

        return "%s" % self.__class__.__module__

    @property
    def manifest(self):
        """Get manifest."""

        return self.context.manager.catalog[self.name]

    @property
    def service_id(self):
        """Get service_id."""

        return self._service_id

    @service_id.setter
    def service_id(self, value):
        """Set service_id."""

        if isinstance(value, uuid.UUID):
            self._service_id = value
        else:
            self._service_id = uuid.UUID(value)

    @property
    def every(self):
        """Return loop period."""

        return self.params["every"]

    @every.setter
    def every(self, value):
        """Set loop period."""

        self.params["every"] = int(value)

        if self.worker:
            self.stop()
            self.start()

    def write_points(self, points):
        """Write points to context."""

        self.context.write_points(points)

    def start(self):
        """Start control loop."""

        # Not supposed to run a loop
        if self.every == -1:
            return

        # Start the control loop
        self.worker = \
            tornado.ioloop.PeriodicCallback(self.loop, self.every)

        self.worker.start()

    def stop(self):
        """Stop control loop."""

        # save state
        self.save_service_state()

        # Not supposed to run a loop
        if self.every == -1:
            return

        # stop the control loop
        self.worker.stop()
        self.worker = None

    def loop(self):
        """Control loop."""

        self.log.info("Empty loop")

    def to_str(self):
        """Return an ASCII representation of the object."""

        return "%s" % self.name

    def __str__(self):
        return self.to_str()

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if isinstance(other, EService):
            return self.name == other.name and self.every == other.every
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return self.__class__.__name__ + "('" + self.to_str() + "')"



#!/usr/bin/env python3
#
# Copyright (c) 2019 Roberto Riggio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Exposes a RESTful interface ."""

import empower_core.apimanager.apimanager as apimanager


# pylint: disable=W0223
class AccountsHandler(apimanager.APIHandler):
    """All the accounts defined in the controller."""

    URLS = [r"/api/v1/accounts/?",
            r"/api/v1/accounts/([a-zA-Z0-9:.]*)/?"]

    @apimanager.validate(max_args=1)
    def get(self, username=None):
        """List the accounts.

        Args:

            [0]: the username

        Example URLs:

            GET /api/v1/accounts

            [
                {
                    "email": "admin@empower.it",
                    "name": "admin",
                    "username": "root"
                },
                {
                    "email": "foo@empower.it",
                    "name": "Foo",
                    "username": "foo"
                },
                {
                    "email": "bar@empower.it",
                    "name": "Bar",
                    "username": "bar"
                }
            ]

            GET /api/v1/accounts/root

            {
                "email": "admin@empower.it",
                "name": "admin",
                "username": "root"
            }
        """

        return self.service.accounts[username] \
            if username else self.service.accounts

    @apimanager.validate(returncode=201)
    def post(self, **kwargs):
        """Create a new account.

        Request:

            version: protocol version (1.0)
            username: username (mandatory)
            password: password (mandatory)
            name: name (mandatory)
            email: email (mandatory)

        Example URLs:

            POST /api/v1/accounts

            {
              "version" : 1.0,
              "username" : "foo",
              "password" : "foo",
              "name" : "foo",
              "email" : "foo@empower.io"
            }
        """

        self.service.create(username=kwargs['username'],
                            password=kwargs['password'],
                            name=kwargs['name'],
                            email=kwargs['email'])

    @apimanager.validate(returncode=204, min_args=1, max_args=1)
    def put(self, *args, **kwargs):
        """Update an account.

        Args:

            [0]: the username

        Request:

            version: protocol version (1.0)
            name: name (mandatory)
            email: email (mandatory)
            password: password (optional)
            new_password: new_password (optional)
            new_password_confirm: new_password_confirm (optional)

        Example URLs:

            PUT /api/v1/accounts/test

            {
              "version" : 1.0,
              "name" : "foo",
              "email" : "foo@empowr.io",
              "new_password" : "new",
              "new_password_confirm" : "new",
            }

            PUT /api/v1/accounts/test

            {
              "version" : 1.0,
              "name" : "foo",
              "email" : "foo@empowr.io",
            }
        """

        username = args[0]
        password = None
        name = kwargs['name'] if 'name' in kwargs else None
        email = kwargs['email'] if 'email' in kwargs else None

        if 'new_password' in kwargs and 'new_password_confirm' in kwargs:

            if kwargs['new_password'] != kwargs['new_password_confirm']:
                raise ValueError("Passwords do not match")

            password = kwargs['new_password']

        self.service.update(username=username, password=password, name=name,
                            email=email)

    @apimanager.validate(returncode=204, min_args=1, max_args=1)
    def delete(self, username):
        """Delete an account.

        Args:

            [0]: the username

        Example URLs:

            DELETE /api/v1/accounts/foo
        """

        self.service.remove(username)



#!/usr/bin/env python3
#
# Copyright (c) 2019 Roberto Riggio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Exposes a RESTful interface ."""

import uuid

import empower_core.apimanager.apimanager as apimanager


# pylint: disable=W0223
class ProjectsHandler(apimanager.APIHandler):
    """Projects handler"""

    URLS = [r"/api/v1/projects/?",
            r"/api/v1/projects/([a-zA-Z0-9-]*)/?"]

    @apimanager.validate(min_args=0, max_args=1)
    def get(self, project_id=None):
        """Lists all the projects.

        Args:

            [0], the project id (optional)

        Example URLs:

            GET /api/v1/projects

            [
                {
                    "bootstrap": {
                    },
                    "desc": "Project description",
                    "owner": "foo",
                    "project_id": "52313ecb-9d00-4b7d-b873-b55d3d9ada26",
                }
            ]

            GET /api/v1/projects/52313ecb-9d00-4b7d-b873-b55d3d9ada26

            {
                "bootstrap": {
                },
                "desc": "Project description",
                "owner": "foo",
                "project_id": "52313ecb-9d00-4b7d-b873-b55d3d9ada26",
            }
        """

        return self.service.projects[uuid.UUID(project_id)] \
            if project_id else self.service.projects

    @apimanager.validate(returncode=201, min_args=0, max_args=1)
    def post(self, *args, **kwargs):
        """Create a new project.

        Args:

            [0], the project id (optional)

        Request:

            version: protocol version (1.0)
            desc: a human-readable description of the project
            owner: the username of the requester
        """

        project_id = uuid.UUID(args[0]) if args else uuid.uuid4()

        project = self.service.create(project_id=project_id,
                                      desc=kwargs['desc'],
                                      owner=kwargs['owner'])

        self.set_header("Location", "/api/v1/projects/%s" % project.project_id)

    @apimanager.validate(returncode=204, min_args=1, max_args=1)
    def put(self, *args, **kwargs):
        """Update a project.

        Args:

            [0], the project id (mandatory)

        Request:

            version: protocol version (1.0)
            desc: a human-readable description of the project
        """

        project_id = uuid.UUID(args[0])

        self.service.update(project_id=project_id, desc=kwargs['desc'])

    @apimanager.validate(returncode=204, min_args=0, max_args=1)
    def delete(self, project_id=None):
        """Delete one or all projects.

        Args:

            [0], the projects id

        Example URLs:

            DELETE /api/v1/projects
            DELETE /api/v1/projects/52313ecb-9d00-4b7d-b873-b55d3d9ada26
        """

        if project_id:
            self.service.remove(uuid.UUID(project_id))
        else:
            self.service.remove_all()



#!/usr/bin/env python3
#
# Copyright (c) 2019 Roberto Riggio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Print help message."""

import argparse
import sys

import empower_core.command as command


def pa_cmd(args, cmd):
    """ Help option parser. """

    usage = "%s <cmd>" % command.USAGE.format(cmd)
    (args, leftovers) = \
        argparse.ArgumentParser(usage=usage).parse_known_args(args)
    return args, leftovers


def do_cmd(gargs, args, leftovers):
    """ Help execute method. """

    if len(leftovers) != 1:
        print("No command specified")
        command.print_available_cmds()
        sys.exit()

    try:
        (parse_args, _) = command.CMDS[leftovers[0]]
        parse_args(['--help'], leftovers[0])
    except KeyError:
        print("Invalid command: %s is an unknown command." % leftovers[0])
        sys.exit()



#!/usr/bin/env python3
#
# Copyright (c) 2019 Roberto Riggio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Access Control List."""

from pymodm import MongoModel, fields

from empower_core.etheraddress import EtherAddressField
from empower_core.serialize import serializable_dict


@serializable_dict
class ACL(MongoModel):
    """Access Control List."""

    addr = EtherAddressField(primary_key=True)
    desc = fields.CharField(required=True)

    def to_dict(self):
        """Return JSON-serializable representation of the object."""

        out = {
            'addr': self.addr,
            'desc': self.desc,
        }

        return out

    def to_str(self):
        """Return an ASCII representation of the object."""

        return "%s" % self.addr

    def __str__(self):
        return self.to_str()

    def __hash__(self):
        return hash(self.addr)

    def __eq__(self, other):
        if isinstance(other, ACL):
            return self.addr == other.addr
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return self.__class__.__name__ + "('" + self.to_str() + "')"



#!/usr/bin/env python3
#
# Copyright (c) 2020 Roberto Riggio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Exposes a RESTful interface ."""

import empower_core.apimanager.apimanager as apimanager


# pylint: disable=W0223
class CatalogHandler(apimanager.APIHandler):
    """Access the applications catalog."""

    URLS = [r"/api/v1/projects/catalog/?",
            r"/api/v1/projects/catalog/(.*)/?"]

    @apimanager.validate(min_args=0, max_args=1)
    def get(self, name=None):
        """List of available applications

        Example URLs:

             GET /api/v1/projects/catalog

            {
                empower.apps.wifimobilitymanager.wifimobilitymanager: {
                    params: {
                        service_id: {
                            desc: "The unique UUID of the application.",
                            mandatory: true,
                            type: "UUID"
                        },
                        project_id: {
                            desc: "The project on which this app must
                                be executed.",
                            mandatory: true,
                            type: "UUID"
                        },
                        every: {
                            desc: "The control loop period (in ms).",
                            mandatory: false,
                            default: 2000,
                            type: "int"
                        }
                    },
                    name:
                        "empower.apps.wifimobilitymanager.wifimobilitymanager",
                    desc: "A simple Wi-Fi mobility manager."
                }
            }
        """

        return self.service.catalog[name] if name else self.service.catalog



#!/usr/bin/env python3
#
# Copyright (c) 2019 Roberto Riggio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Wi-Fi Service Set Identifier (SSID)."""

import re

from pymodm.errors import ValidationError
from pymodm.base.fields import MongoBaseField
from empower_core.serialize import serializable_string

WIFI_NWID_MAXSIZE = 32


@serializable_string
class SSID:
    """Wi-Fi Service Set Identifier (SSID)."""

    def __init__(self, ssid=None):

        if not ssid:
            ssid = ""

        if isinstance(ssid, bytes):
            self.ssid = ssid.decode('UTF-8').rstrip('\0')
        elif isinstance(ssid, str):
            allowed = re.compile(r'^[a-zA-Z0-9 ]*$',
                                 re.VERBOSE | re.IGNORECASE)
            if allowed.match(ssid) is None:
                raise ValueError("Invalid SSID name")
            self.ssid = ssid
        elif isinstance(ssid, SSID):
            self.ssid = str(ssid)
        else:
            raise ValueError("SSID must be a string or an array of UTF-8 "
                             "encoded bytes array of UTF-8 encoded bytes")

    def to_raw(self):
        """ Return the bytes represenation of the SSID """

        bytes_ssid = self.ssid.encode('UTF-8')
        return bytes_ssid + b'\0' * (WIFI_NWID_MAXSIZE + 1 - len(bytes_ssid))

    def to_str(self):
        """Return an ASCII representation of the object."""

        return self.ssid

    def __bool__(self):
        return bool(self.ssid)

    def __str__(self):
        return self.to_str()

    def __len__(self):
        return len(self.ssid)

    def __hash__(self):
        return hash(self.ssid)

    def __eq__(self, other):
        if isinstance(other, SSID):
            return self.ssid == other.ssid
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return self.__class__.__name__ + "('" + self.to_str() + "')"


class SSIDField(MongoBaseField):
    """A field that stores SSIDs."""

    def __init__(self, verbose_name=None, mongo_name=None, **kwargs):

        super(SSIDField, self).__init__(verbose_name=verbose_name,
                                        mongo_name=mongo_name,
                                        **kwargs)

        def validate_ssid(value):

            try:
                SSID(value)
            except ValueError:
                msg = '%r is not a valid SSID.' % value
                raise ValidationError(msg)

        self.validators.append(validate_ssid)

    @classmethod
    def to_mongo(cls, value):
        """Convert value for storage."""

        try:
            return str(value)
        except ValueError:
            msg = '%r is not a valid SSID.' % value
            raise ValidationError(msg)

    @classmethod
    def to_python(cls, value):
        """Convert value back to Python."""

        try:
            return SSID(value)
        except ValueError:
            msg = '%r is not a valid SSID.' % value
            raise ValidationError(msg)



#!/usr/bin/env python3
#
# Copyright (c) 2019 Roberto Riggio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Exposes a RESTful interface ."""

import uuid

import empower_core.apimanager.apimanager as apimanager


# pylint: disable=W0223
class AppCallbacksHandler(apimanager.APIHandler):
    """Apps handler."""

    URLS = [
        r"/api/v1/projects/([a-zA-Z0-9-]*)/apps/([a-zA-Z0-9-]*)/callbacks/?",
        r"/api/v1/projects/([a-zA-Z0-9-]*)/apps/([a-zA-Z0-9-]*)/callbacks/"
        "([a-z]*)/?"
    ]

    @apimanager.validate(min_args=2, max_args=3)
    def get(self, project_id, service_id, callback=None):
        """List the callback.

        Args:

            [0]: the project id (mandatory)
            [1]: the app id (optional)
            [2]: the callback (optional)

        Example URLs:

            GET /api/v1/projects/52313ecb-9d00-4b7d-b873-b55d3d9ada26/apps/
                0f91e8ad-1c2a-4b06-97f9-e34097c4c1d0/callbacks
            {
                default: {
                    type: "url",
                    name: "default",
                    callback: "http://www.domain.io/resource"
                }
            }

            GET /api/v1/projects/52313ecb-9d00-4b7d-b873-b55d3d9ada26/apps/
                0f91e8ad-1c2a-4b06-97f9-e34097c4c1d0/callbacks/default

            {
                type: "url",
                name: "default",
                callback: "http://www.domain.io/resource"
            }
        """

        project_id = uuid.UUID(project_id)
        project = self.service.projects[project_id]

        service_id = uuid.UUID(service_id)
        service = project.services[service_id]

        return service.callbacks[callback] if callback else service.callbacks

    @apimanager.validate(returncode=201, min_args=2, max_args=2)
    def post(self, *args, **kwargs):
        """Add a callback.

        Args:

            [0]: the project id (mandatory)
            [1]: the app id (mandatory)

        Request:

            version: protocol version (1.0)
            name: the name of the callback (mandatory)
            callback: the callback URL (mandatory)

        Example URLs:

            POST /api/v1/workers
            {
                "version": "1.0",
                "name": "default",
                "callback_type": "rest"
                "callback": "http://www.domain.io/resource"
            }
        """

        project_id = uuid.UUID(args[0])
        project = self.service.projects[project_id]

        service_id = uuid.UUID(args[1])
        service = project.services[service_id]
        service.add_callback(name=kwargs['name'],
                             callback_type=kwargs['callback_type'],
                             callback=kwargs['callback'])

        self.set_header("Location",
                        "/api/v1/projects/%s/apps/%s/callback/%s" %
                        (project_id, service_id, kwargs['name']))

    @apimanager.validate(returncode=204, min_args=3, max_args=3)
    def delete(self, project_id, service_id, callback):
        """Stop a worker.

        Args:

            [0]: the project id (mandatory)
            [1]: the app id (mandatory)
            [2]: the callback (mandatory)

        Example URLs:

            DELETE /api/v1/workers/08e14f40-6ebf-47a0-8baa-11d7f44cc228/default
        """

        project_id = uuid.UUID(project_id)
        project = self.service.projects[project_id]

        service_id = uuid.UUID(service_id)
        service = project.services[service_id]

        service.rem_callback(callback)



#!/usr/bin/env python3
#
# Copyright (c) 2019 Roberto Riggio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Project Class."""

from importlib import import_module

from pymodm import fields

from empower_core.envmanager.env import Env
from empower_core.launcher import srv_or_die
from empower_core.serialize import serializable_dict
from empower_core.app import EApp


@serializable_dict
class Project(Env):
    """Project class.

    Attributes:
        owner: The username of the user that requested this pool
        desc: A human readable description of the project
    """

    owner = fields.CharField(required=True)
    desc = fields.CharField(required=True)

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        # Save pointer to ProjectManager
        self.manager = srv_or_die("projectsmanager")

    def load_service(self, service_id, name, params):
        """Load a service instance."""

        init_method = getattr(import_module(name), "launch")
        service = init_method(context=self, service_id=service_id, **params)

        if not isinstance(service, EApp):
            raise ValueError("Service %s not EApp type" % name)

        return service

    def to_dict(self):
        """Return JSON-serializable representation of the object."""

        output = super().to_dict()

        output['owner'] = self.owner
        output['desc'] = self.desc

        return output



#!/usr/bin/env python3
#
# Copyright (c) 2019 Roberto Riggio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Setup script."""

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="empower-core",
    version="1.0.10",
    author="Roberto Riggio",
    author_email="roberto.riggio@gmail.com",
    description="The EmPOWER core library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://5g-empower.github.io/",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)



#!/usr/bin/env python3
#
# Copyright (c) 2019 Roberto Riggio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""API Manager (REST Northbound Interface)."""

import inspect
import json
import base64
import re

from uuid import UUID

import tornado.web
import tornado.httpserver

from tornado.web import Application
from pymodm.errors import ValidationError

from empower_core.serialize import serialize
from empower_core.service import EService
from empower_core.launcher import srv_or_die, srv

DEBUG = True
DEFAULT_PORT = 8888
DEFAULT_WEBUI = "/var/www/webui/"
COOKIE_SECRET = b'xyRTvZpRSUyk8/9/McQAvsQPB4Rqv0w9mBtIpH9lf1o='
LOGIN_URL = "/auth/login"


def validate(returncode=200, min_args=0, max_args=0):
    """Validate REST method."""

    def decorator(func):

        def magic(self, *args):

            try:

                if len(args) < min_args or len(args) > max_args:
                    msg = "Invalid url (%u, %u)" % (min_args, max_args)
                    raise ValueError(msg)

                params = {}

                if self.request.body and json.loads(self.request.body):
                    params = json.loads(self.request.body)

                if "version" in params:
                    del params["version"]

                output = func(self, *args, **params)

                if returncode == 200:
                    self.write_as_json(output)

            except KeyError as ex:
                self.send_error(404, message=str(ex))

            except ValueError as ex:
                self.send_error(400, message=str(ex))

            except AttributeError as ex:
                self.send_error(400, message=str(ex))

            except TypeError as ex:
                self.send_error(400, message=str(ex))

            except ValidationError as ex:
                self.send_error(400, message=ex.message)

            self.set_status(returncode, None)

        magic.__doc__ = func.__doc__

        return magic

    return decorator


# pylint: disable=W0223
class BaseHandler(tornado.web.RequestHandler):
    """Base Handler."""

    # service associated to this handler
    service = None

    URLS = []

    def get_current_user(self):
        """Return username of the currently logged user or None."""

        return self.get_secure_cookie("username")

    @classmethod
    def auth_based(cls):
        """Return true if both account and project managers are available"""

        pmngr = srv("projectsmanager")
        amngr = srv("accountsmanager")

        return bool(pmngr and amngr)


class IndexHandler(BaseHandler):
    """Index page handler."""

    URLS = [r"/", r"/([a-z]*).html"]

    def get_project(self):
        """Get the current project or return None if not project is set."""

        # check if a project has been selected
        project_id = self.get_secure_cookie("project_id")

        if not project_id:
            return None

        project_id = UUID(project_id.decode('UTF-8'))
        projects_manager = srv_or_die("projectsmanager")

        if project_id not in projects_manager.projects:
            self.clear_cookie("project_id")
            return None

        return projects_manager.projects[project_id]

    @tornado.web.authenticated
    def get(self, args=None):
        """Render index page."""

        try:

            if self.auth_based():

                username = self.get_secure_cookie("username").decode('UTF-8')
                accounts_manager = srv_or_die("accountsmanager")
                account = accounts_manager.accounts[username]

                page = "index.html" if not args else "%s.html" % args

                self.render(page,
                            username=account.username,
                            password=account.password,
                            name=account.name,
                            email=account.email,
                            project=self.get_project())

            else:

                page = "index.html" if not args else "%s.html" % args

                self.render(page)

        except KeyError as ex:
            self.send_error(404, message=str(ex))

        except ValueError as ex:
            self.send_error(400, message=str(ex))


class AuthSwitchProjectHandler(BaseHandler):
    """Login page handler."""

    URLS = [r"/auth/switch_project"]

    def get(self):
        """Set the active project."""

        username = self.get_secure_cookie("username").decode('UTF-8')

        # if root deselect project
        if username == "root":
            self.clear_cookie("project_id")
            self.redirect('/')
            return

        # check if the project id is in the URL
        project_id = self.get_argument("project_id", None)

        # reset project selected
        if not project_id:
            self.clear_cookie("project_id")
            self.redirect('/')
            return

        try:

            # set project
            project_id = UUID(project_id)
            projects_manager = srv_or_die("projectsmanager")
            project = projects_manager.projects[project_id]

            if project.owner != username:
                self.clear_cookie("project_id")
                self.redirect('/')
                return

            self.set_secure_cookie("project_id", str(project.project_id))

        except KeyError:
            self.clear_cookie("project_id")

        except ValueError:
            self.clear_cookie("project_id")

        self.redirect('/')


class AuthLoginHandler(BaseHandler):
    """Login page handler."""

    URLS = [r"/auth/login"]

    def get(self):
        """Render login page."""

        if not self.auth_based():
            self.set_secure_cookie("username", "none")
            self.redirect('/')
            return

        if self.get_current_user():
            self.redirect('/')
            return

        self.render("login.html", error=self.get_argument("error", ""))

    def post(self):
        """Process login credentials."""

        username = self.get_argument("username", "")
        password = self.get_argument("password", "")

        accounts_manager = srv_or_die("accountsmanager")

        if accounts_manager.check_permission(username, password):
            self.set_secure_cookie("username", username)
            self.redirect("/index.html")
        else:
            self.clear_cookie("username")
            self.redirect("/auth/login?error=Wrong credentials!")


class AuthLogoutHandler(BaseHandler):
    """Logout page handler."""

    URLS = [r"/auth/logout"]

    def get(self):
        """Process logout request."""

        self.clear_cookie("username")
        self.clear_cookie("project_id")
        self.redirect("/auth/login")


class APIHandler(tornado.web.RequestHandler):
    """Base class for all the REST calls."""

    # service associated to this handler
    service = None

    @classmethod
    def auth_based(cls):
        """Return true if both account and project managers are available"""

        pmngr = srv("projectsmanager")
        amngr = srv("accountsmanager")

        return bool(pmngr and amngr)

    def write_error(self, status_code, **kwargs):
        """Write error as JSON message."""

        self.set_header('Content-Type', 'application/json')

        value = {
            "title": self._reason,
            "status_code": status_code,
            "detail": kwargs.get("message"),
        }

        self.finish(json.dumps(serialize(value), indent=4))

    def write_as_json(self, value):
        """Return reply as a json document."""

        self.write(json.dumps(serialize(value), indent=4))

    def prepare(self):
        """Prepare to handler reply."""

        self.set_header('Content-Type', 'application/json')

        # no account manager or project manager
        if not self.auth_based():
            return

        # get requests do not require authentication
        if self.request.method == "GET":
            return

        accounts_manager = srv_or_die("accountsmanager")
        projects_manager = srv_or_die("projectsmanager")

        auth_header = self.request.headers.get('Authorization')

        if auth_header is None or not auth_header.startswith('Basic '):
            self.set_header('WWW-Authenticate', 'Basic realm=Restricted')
            self.send_error(401, message="Missing authorization header")
            return

        auth_bytes = bytes(auth_header[6:], 'utf-8')
        auth_decoded = base64.b64decode(auth_bytes).decode()
        username, password = auth_decoded.split(':', 2)

        # account does not exists
        if not accounts_manager.check_permission(username, password):
            self.send_error(401, message="Invalid username/password")
            return

        account = accounts_manager.accounts[username]

        # root can do everything
        if account.username == "root":
            return

        # check if logged user is accessing his/her own account
        if self.request.uri.startswith("/api/v1/accounts"):

            pattern = re.compile("/api/v1/accounts/([a-zA-Z0-9:-]*)/?")
            match = pattern.match(self.request.uri)

            if match and match.group(1):
                username = match.group(1)
                if username == account.username:
                    return

        # check if logged user is accessing one of his/her projects
        if self.request.uri.startswith("/api/v1/projects"):

            pattern = re.compile("/api/v1/projects/([a-zA-Z0-9-]*)/?")
            match = pattern.match(self.request.uri)

            if match and match.group(1):
                project_id = UUID(match.group(1))
                if project_id in projects_manager.projects:
                    project = projects_manager.projects[project_id]
                    if account.username == project.owner:
                        return

        self.send_error(401, message="URI not authorized")


BOILER_PLATE = """# REST API

The REST API consists of a set of RESTful resources and their attributes.
The base URL for the REST API is the following:

    http{s}://{username}:{password}@{hostname}:{port}/api/v1/{resource}

Of course, you need to replace hostname and port with the hostname/port
combination for your controller.

The current (and only) version of the API is v1.

The REST API uses HTTP basic authentication control access to RESTful resource.

Notice that there are two kinds of accounts:

 * user accounts, which have complete CRUD access only to all the URLs that
 begins with /api/v1/projects/{project_id}.

 * root account, which has complete CRUD access to all URLs. All the URLs that
 DO NOT start with /api/v1/projects/{project_id} require a root account to
 be accessed. The only exception is the URL /api/v1/accounts/{user_id} which
 is fully accessible to all users.
 """


# pylint: disable=W0223
class DocHandler(APIHandler):
    """Generates markdown documentation."""

    URLS = [r"/api/v1/doc/?"]

    def get(self):
        """Generates markdown documentation.

        Args:

            None

        Example URLs:

            GET /api/v1/doc
        """

        exclude_list = ["StaticFileHandler", "DocHandler"]
        handlers = set()
        accum = [BOILER_PLATE]

        for rule in self.service.application.default_router.rules:
            if not rule.target.rules:
                continue
            handlers.add(rule.target.rules[0].target)

        handlers = sorted(handlers, key=lambda x: x.__name__)

        accum.append("## <a name='handlers'></a>Handlers\n")

        for handler in handlers:

            if handler.__name__ in exclude_list:
                continue

            accum.append(" * [%s](#%s)" %
                         (handler.__name__, handler.__name__))

        accum.append("\n")

        for handler in handlers:

            if handler.__name__ in exclude_list:
                continue

            accum.append("# <a name='%s'></a>%s ([Top](#handlers))\n" %
                         (handler.__name__, handler.__name__))

            accum.append("%s\n" % inspect.getdoc(handler))

            if hasattr(handler, "URLS") and handler.URLS:
                accum.append("### URLs\n")
                for url in handler.URLS:
                    accum.append("    %s" % url)

            accum.append("\n")

            if hasattr(handler, "get"):
                doc = inspect.getdoc(getattr(handler, "get"))
                if doc:
                    accum.append("### GET\n")
                    accum.append(doc)
                    accum.append("\n")

            if hasattr(handler, "put"):
                doc = inspect.getdoc(getattr(handler, "put"))
                if doc:
                    accum.append("### PUT\n")
                    accum.append(doc)
                    accum.append("\n")

            if hasattr(handler, "post"):
                doc = inspect.getdoc(getattr(handler, "post"))
                if doc:
                    accum.append("### POST\n")
                    accum.append(doc)
                    accum.append("\n")

            if hasattr(handler, "delete"):
                doc = inspect.getdoc(getattr(handler, "delete"))
                if doc:
                    accum.append("### DELETE\n")
                    accum.append(doc)
                    accum.append("\n")

        self.write('\n'.join(accum))


class APIManager(EService):
    """Service exposing a REST API

    Parameters:
        port: the port on which the HTTP server should listen (optional,
            default: 8888)
    """

    HANDLERS = [IndexHandler, AuthLoginHandler, AuthLogoutHandler,
                DocHandler, AuthSwitchProjectHandler]

    def __init__(self, context, service_id, webui, port):

        super().__init__(context=context, service_id=service_id, webui=webui,
                         port=port)

        self.settings = {
            "static_path": self.webui + "static/",
            "cookie_secret": COOKIE_SECRET,
            "template_path": self.webui + "templates/",
            "login_url": LOGIN_URL,
            "debug": DEBUG,
        }

        self.application = Application([], **self.settings)

        self.http_server = tornado.httpserver.HTTPServer(self.application)

    @property
    def webui(self):
        """Return path to Web UI."""

        return self.params["webui"]

    @webui.setter
    def webui(self, value):
        """Set path to Web UI."""

        if "webui" in self.params and self.params["webui"]:
            raise ValueError("Param webui can not be changed")

        self.params["webui"] = value

    @property
    def port(self):
        """Return port."""

        return self.params["port"]

    @port.setter
    def port(self, value):
        """Set port."""

        if "port" in self.params and self.params["port"]:
            raise ValueError("Param port can not be changed")

        self.params["port"] = int(value)

    def start(self):
        """Start api manager."""

        super().start()

        self.http_server.listen(self.port)

        self.log.info("Listening on port %u", self.port)

        self.http_server.start()

    def register_handler(self, handler):
        """Add a new handler class."""

        for url in handler.URLS:
            self.log.info("Registering URL: %s", url)
            self.application.add_handlers(r".*$", [(url, handler)])


def launch(context, service_id, webui=DEFAULT_WEBUI, port=DEFAULT_PORT):
    """ Initialize the module. """

    return APIManager(context=context, service_id=service_id, webui=webui,
                      port=port)



#!/usr/bin/env python3
#
# Copyright (c) 2019 Roberto Riggio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Launcher module."""

import configparser
import ssl
import uuid
import os
import logging
import logging.config
import sys

from importlib import import_module
from argparse import ArgumentParser

import tornado.ioloop

from pymodm.connection import connect

SERVICES = dict()


def srv(name):
    """Get a service instance or return None"""

    if name in SERVICES:
        return SERVICES[name]

    return None


def srv_or_die(name):
    """Get a service instance or return None"""

    if name in SERVICES:
        return SERVICES[name]

    logging.error("Unable to find service %s", name)

    sys.exit(1)


def _do_launch(managers, managers_order):
    """Parse arguments and launch controller."""

    # Register managers
    for manager in managers_order:

        module = managers[manager]['module']
        params = managers[manager]['params']

        if manager in SERVICES:
            logging.error("%s manager already registered", manager)
            return False

        init_method = getattr(import_module(module), "launch")

        logging.info("Loading manager: %s", manager)

        if params:
            logging.info("  - params: %s", params)

        service = init_method(context=None, service_id=uuid.uuid4(), **params)

        SERVICES[manager] = service

    # Start managers
    for manager in managers_order:

        service = SERVICES[manager]

        logging.info("Starting manager: %s", manager)

        # Register handlers for this services
        api_manager = srv_or_die("apimanager")
        for handler in service.HANDLERS:
            api_manager.register_handler(handler)
            handler.service = service

        # start service
        service.start()

    return True


def _setup_db(args):
    """ Setup db connection. """

    runtime_config = args.config + "/runtime.cfg"
    config = configparser.ConfigParser()
    config.read(runtime_config)

    mongodb_uri = config.get('general', 'mongodb', fallback=None)

    if mongodb_uri:
        connect(mongodb_uri, ssl_cert_reqs=ssl.CERT_NONE)


def _setup_logging(args):
    """ Setup logging. """

    runtime_config = args.config + "/runtime.cfg"
    config = configparser.ConfigParser()
    config.read(runtime_config)

    log_config = config.get('general', 'logging',
                            fallback="/etc/empower/logging.cfg")

    if not os.path.exists(log_config):
        print("Could not find logging config file: %s" % log_config)
        sys.exit(1)

    logging.config.fileConfig(log_config, disable_existing_loggers=False)


def _pre_startup(args):
    """Perform pre-startup operations."""

    _setup_logging(args)
    _setup_db(args)


def _post_startup():
    """Perform post-startup operation."""


def _read_config(args):
    """Read config file."""

    runtime_config = args.config + "/runtime.cfg"
    config = configparser.ConfigParser()
    config.read(runtime_config)

    managers = {}
    managers_order = []

    mngrs = config.get('general', 'managers', fallback=None)

    if not mngrs:
        return managers, managers_order

    for mngr in mngrs.split(","):

        module = config.get(mngr, 'module', fallback=None)

        if not module:
            continue

        managers_order.append(mngr)

        managers[mngr] = {
            "module": module,
            "params": {}
        }

        for param in config[mngr]:

            if param == 'module':
                continue

            managers[mngr]["params"][param] = config[mngr][param]

    return managers, managers_order


def _parse_global_args(config):
    """ Parse global arguments list. """

    parser = ArgumentParser()

    parser.add_argument("-c", "--config",
                        dest="config",
                        default=config,
                        help="Configuration directory, default: %s" %
                        config)

    return parser.parse_known_args(sys.argv[1:])


def main(config=""):
    """Parses the command line and loads the plugins."""

    # parse command line
    args, _ = _parse_global_args(config)

    # perform pre-startup operation
    _pre_startup(args)

    # load managers
    managers, managers_order = _read_config(args)

    # launch managers
    if _do_launch(managers, managers_order):
        _post_startup()
    else:
        raise RuntimeError()

    # start tornado loop
    tornado.ioloop.IOLoop.instance().start()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# Copyright (c) 2019 Roberto Riggio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Exposes a RESTful interface ."""

import empower_core.apimanager.apimanager as apimanager


# pylint: disable=W0223
class EnvHandler(apimanager.APIHandler):
    """Access the system services."""

    URLS = [r"/api/v1/env/?"]

    @apimanager.validate(min_args=0, max_args=1)
    def get(self):
        """Get environment

        Example URLs:

             GET /api/v1/env

            {
                "empower.managers.envmanager.envmanager": {
                    "name": "empower.managers.envmanager.envmanager",
                    "params": {
                        "service_id": "b7d872a2-fee1-442e-b6d9-f33e5ce9fca1",
                        "every": -1
                    }
                }
            }
        """

        return self.service.env



#!/usr/bin/env python3
#
# Copyright (c) 2019 Roberto Riggio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Time series manager."""

from concurrent.futures import ThreadPoolExecutor

from tornado import gen
from influxdb import InfluxDBClient

from empower_core.service import EService
from empower_core.serialize import serialize

DEFAULT_DATABASE = "tsmanager"
DEFAULT_HOST = "localhost"
DEFAULT_PORT = 8086
DEFAULT_USERNAME = "root"
DEFAULT_PASSWORD = "password"


class InfluxTimeSeriesManager(EService):
    """Time series manager."""

    def __init__(self, context, service_id, database, host, port, username,
                 password):

        super().__init__(context=context, service_id=service_id,
                         database=database, host=host, port=port,
                         username=username, password=password)

        self.thread_pool = None
        self.influxdb_client = None

        # stats buffer, cannot rely on ThreadPoolExecutor because
        # it cannot clear pending jobs when shutdown is called
        self.stats = []
        self.busy = False

    def start(self):
        """Start time series manager manager."""

        super().start()

        self.thread_pool = ThreadPoolExecutor(1)
        self.influxdb_client = InfluxDBClient(host=self.host,
                                              port=self.port,
                                              username=self.username,
                                              password=self.password,
                                              timeout=3,
                                              database=self.database)

        try:
            # create database, it has no effect if it is already present
            self.influxdb_client.create_database(self.database)
            self.log.info("Connected to InfluxDB database %s", self.database)
        except Exception as ex:
            self.log.exception(ex)

    @property
    def database(self):
        """Return database."""

        return self.params["database"]

    @database.setter
    def database(self, value):
        """Set database."""

        self.params["database"] = value

    @property
    def host(self):
        """Return host."""

        return self.params["host"]

    @host.setter
    def host(self, value):
        """Set host."""

        self.params["host"] = value

    @property
    def port(self):
        """Return port."""

        return self.params["port"]

    @port.setter
    def port(self, value):
        """Set port."""

        self.params["port"] = int(value)

    @property
    def username(self):
        """Return username."""

        return self.params["username"]

    @username.setter
    def username(self, value):
        """Set username."""

        self.params["username"] = value

    @property
    def password(self):
        """Return password."""

        return self.params["password"]

    @password.setter
    def password(self, value):
        """Set password."""

        self.params["password"] = value

    @gen.coroutine
    def write_points(self, points):
        """Add new points to the DB."""

        # the sender thread is already working, buffer data
        if self.busy:
            self.stats.append(points)
            return

        self.busy = True

        error = yield self.thread_pool.submit(self.__write_points_worker,
                                              points)
        self.busy = False

        # clear buffer in case of error
        if error:
            self.stats.clear()

        # pop buffered data and send it
        if self.stats:
            self.write_points(self.stats.pop(0))

    def __write_points_worker(self, points):

        try:
            self.influxdb_client.write_points(points=serialize(points))
        except Exception as ex:
            self.log.exception(ex)
            return True

        return False


def launch(context, service_id, database=DEFAULT_DATABASE,
           host=DEFAULT_HOST, port=DEFAULT_PORT, username=DEFAULT_USERNAME,
           password=DEFAULT_PASSWORD):
    """ Initialize the module. """

    return InfluxTimeSeriesManager(context=context, service_id=service_id,
                                   database=database, host=host, port=port,
                                   username=username, password=password)



#!/usr/bin/env python3
#
# Copyright (c) 2019 Roberto Riggio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Exposes a RESTful interface ."""

import uuid

import empower_core.apimanager.apimanager as apimanager


# pylint: disable=W0223
class AppsHandler(apimanager.APIHandler):
    """Applications handler."""

    URLS = [r"/api/v1/projects/([a-zA-Z0-9-]*)/apps/?",
            r"/api/v1/projects/([a-zA-Z0-9-]*)/apps/([a-zA-Z0-9-]*)/?"]

    @apimanager.validate(min_args=1, max_args=2)
    def get(self, project_id, app_id=None):
        """List the apps.

        Args:

            [0]: the project id (mandatory)
            [1]: the app id (optional)

        Example URLs:

            GET /api/v1/projects/52313ecb-9d00-4b7d-b873-b55d3d9ada26/apps

            [
                {
                    "counters": {},
                    "name":
                        "empower.apps.wifimobilitymanager.wifimobilitymanager",
                    "params": {
                        "every": 2000,
                        "project_id": "52313ecb-9d00-4b7d-b873-b55d3d9ada26",
                        "service_id": "7069c865-8849-4840-9d96-e028663a5dcf"
                    },
                    "stats": {
                        "last_run": "2019-08-23 09:45:20.234651"
                    }
                }
            ]

            GET /api/v1/projects/52313ecb-9d00-4b7d-b873-b55d3d9ada26/apps/
                7069c865-8849-4840-9d96-e028663a5dcf

            {
                "counters": {},
                "name": "empower.apps.wifimobilitymanager.wifimobilitymanager",
                "params": {
                    "every": 2000,
                    "project_id": "52313ecb-9d00-4b7d-b873-b55d3d9ada26",
                    "service_id": "7069c865-8849-4840-9d96-e028663a5dcf"
                },
                "stats": {
                    "last_run": "2019-08-23 09:47:04.361268"
                }
            }
        """

        project_id = uuid.UUID(project_id)
        project = self.service.projects[project_id]

        return project.services[uuid.UUID(app_id)] \
            if app_id else project.services

    @apimanager.validate(returncode=201, min_args=1, max_args=2)
    def post(self, *args, **kwargs):
        """Start a new app.

        Args:

            [0]: the project id (mandatory)

        Request:

            version: protocol version (1.0)
            params: the list of parmeters to be set

        Example URLs:

            POST /api/v1/projects/52313ecb-9d00-4b7d-b873-b55d3d9ada26/apps

            {
                "version": "1.0",
                "name": "empower.apps.wifimobilitymanager.wifimobilitymanager",
                "params": {
                    "every": 5000
                }
            }
        """

        project_id = uuid.UUID(args[0])
        project = self.service.projects[project_id]

        service_id = uuid.UUID(args[1]) if len(args) > 1 else None

        params = kwargs['params'] if 'params' in kwargs else {}

        service = project.register_service(name=kwargs['name'],
                                           params=params,
                                           service_id=service_id)

        self.set_header("Location", "/api/v1/projects/%s/apps/%s" %
                        (project.project_id, service.service_id))

    @apimanager.validate(returncode=204, min_args=2, max_args=2)
    def put(self, *args, **kwargs):
        """Update the configuration of an applications.

        Args:

            [0]: the project id (mandatory)
            [1]: the app id (mandatory)

        Request:

            version: protocol version (1.0)
            params: the list of parmeters to be set

        Example URLs:

            PUT /api/v1/projects/52313ecb-9d00-4b7d-b873-b55d3d9ada26/apps/
                7069c865-8849-4840-9d96-e028663a5dcf

            {
                "version": "1.0",
                "params": {
                    "every": 5000
                }
            }
        """

        project_id = uuid.UUID(args[0])
        project = self.service.projects[project_id]
        service_id = uuid.UUID(args[1])
        params = kwargs['params'] if 'params' in kwargs else {}

        project.reconfigure_service(service_id, params)

    @apimanager.validate(returncode=204, min_args=2, max_args=2)
    def delete(self, project_id, app_id):
        """Stop an app.

        Args:

            [0]: the project id (mandatory)
            [1]: the app id (mandatory)

        Example URLs:

            DELETE /api/v1/projects/52313ecb-9d00-4b7d-b873-b55d3d9ada26/apps/
                7069c865-8849-4840-9d96-e028663a5dcf
        """

        project_id = uuid.UUID(project_id)
        project = self.service.projects[project_id]

        app_id = uuid.UUID(app_id)

        project.unregister_service(app_id)



#!/usr/bin/env python3
#
# Copyright (c) 2019 Roberto Riggio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""JSON Serializer."""

import uuid
import ipaddress
import datetime
import types

from functools import singledispatch


@singledispatch
def serialize(obj):
    """Recursively serialise objects."""

    return obj


@serialize.register(types.FunctionType)
@serialize.register(types.MethodType)
def _(obj):
    return obj.__name__


@serialize.register(dict)
def _(obj):
    return {str(k): serialize(v) for k, v in obj.items()}


@serialize.register(list)
@serialize.register(set)
@serialize.register(tuple)
def _(obj):
    return [serialize(v) for v in obj]


@serialize.register(datetime.datetime)
@serialize.register(uuid.UUID)
@serialize.register(ipaddress.IPv4Address)
def _(obj):
    return str(obj)


def serializable_string(cls):
    """Decorator for classes that can be serialized as dicts."""

    def decorator(cls):

        @serialize.register(cls)
        def _(obj):
            return str(obj)

        return cls

    return decorator(cls)


def serializable_dict(cls):
    """Decorator for classes that can be serialized as dicts."""

    def decorator(cls):

        @serialize.register(cls)
        def _(obj):
            return serialize(obj.to_dict())

        return cls

    return decorator(cls)
